/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Clack.
 * 
 * Clack is free software, also known as "open source."  You can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * (GPL), version 2, as published by the Free Software Foundation (FSF).
 * 
 * Clack is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with Clack; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 * 
 * Clack is modelled on and includes some code fragments from MIT's Click
 * system (http://www.pdos.lcs.mit.edu/click/) which is copyright (c)
 * 1999-2000 Massachusetts Institute of Technology and also GPL'ed.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <clack/c/packet.h>
#include <clack/c/packet_ip.h>
#include <clack/c/print_packet.h>
#include <clack/c/lookuproute.h>
#include <arpa/inet.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/*
 * Imports:  rtab   - The routing table in string format.
 *           ipoutN - Outgoing links that correspond to last field in table.
 */
extern clack_error_t		push_ipout0(clack_packet_t *);
extern clack_error_t		push_ipout1(clack_packet_t *);
extern clack_error_t		push_ipout2(clack_packet_t *);
extern struct rtab_descriptor	route_descriptor_table[];

/*
 * Exports:  ipin   - Incoming IP packets to be routed.
 */
clack_error_t			push(clack_packet_t *);

/*
 * Internal format of the table. Simple linear search!
 */
struct rtab_entry {
	struct in_addr	dst, mask, gateway;
	clack_error_t	(*push)(clack_packet_t *);
};
static struct rtab_entry	*myroutetab;
static int			myroutetab_count;

/*
 * Cache last route.
 */
static struct in_addr		last_addr1, last_addr2;
static int			last_best1, last_best2;

clack_error_t
init(void)
{
	int			index, count;
	struct rtab_descriptor	*rtabp;

	/*
	 * Count up table entries.
	 */
	count = 0;
	rtabp = route_descriptor_table;
	while (rtabp->dst) {
		count++;
		rtabp++;
	}

	/*
	 * Allocate our internal table.
	 */
	myroutetab_count = count;
	myroutetab = (struct rtab_entry *)
		     malloc(count * sizeof(struct rtab_entry));
	assert(myroutetab);

	/*
	 * Convert all the strings to internal format.
	 */
	rtabp = route_descriptor_table;
	index = 0;
	while (rtabp->dst) {
		inet_aton(rtabp->dst,     &myroutetab[index].dst);
		inet_aton(rtabp->mask,    &myroutetab[index].mask);
		inet_aton(rtabp->gateway, &myroutetab[index].gateway);

		/*
		 * Stash the function pointer so we don't have to do this
		 * calculation later.
		 */
		switch (rtabp->link) {
		case 0:
			myroutetab[index].push = push_ipout0;
			break;
		case 1:
			myroutetab[index].push = push_ipout1;
			break;
		case 2:
			myroutetab[index].push = push_ipout2;
			break;
		default:
			printf("LookupRoute(init): Something went wrong!");
		}
		rtabp++;
		index++;
	}
	
	return 0;
}

/*
 * Find the route and push the packet out the corresponding link.
 */
clack_error_t
push(clack_packet_t *p)
{
	int			i, best = -1;

	/*
	 * Check cached routes first.
	 */
	if (p->anno.next_hop.s_addr == last_addr1.s_addr) {
		struct rtab_entry	*mapping = &myroutetab[last_best1];

		if (mapping->gateway.s_addr)
			p->anno.next_hop.s_addr = mapping->gateway.s_addr;
		
		myroutetab[last_best1].push(p);
		return 0;
	}
	else if (p->anno.next_hop.s_addr == last_addr2.s_addr) {
		struct rtab_entry	*mapping = &myroutetab[last_best2];

		if (mapping->gateway.s_addr)
			p->anno.next_hop.s_addr = mapping->gateway.s_addr;
		
		myroutetab[last_best2].push(p);
		return 0;
	}
	
	/*
	 * Scan the table. Lookup is by longest prefix.
	 */
	for (i = 0; i < myroutetab_count; i++) {
		if ((p->anno.next_hop.s_addr & myroutetab[i].mask.s_addr) ==
		    myroutetab[i].dst.s_addr) {
			if (best == -1 || ~myroutetab[i].mask.s_addr <
			                  myroutetab[best].mask.s_addr) {
				best = i;
			}
		}
	}
	if (best == -1) {
		printf("LookupRoute(push): No route for ");
		print_IP2(&p->anno.next_hop);
		printf("\n");
		clack_packet_release(p);
		return 0;
	}

	/*
	 * Set cached entries.
	 */
	last_addr2 = last_addr1;
	last_best2 = last_best1;
	last_addr1 = p->anno.next_hop;
	last_best1 = best;

	/*
	 * Shoot the packet out the appropriate link
	 */
	if (myroutetab[best].gateway.s_addr)
		p->anno.next_hop.s_addr = myroutetab[best].gateway.s_addr;
	myroutetab[best].push(p);
	
	return 0;
}
